package com.tagfriendly.solr;

import com.tagfriendly.relational.Artist;
import com.tagfriendly.relational.Album;
import com.tagfriendly.relational.Track;

import java.util.List;
import java.util.ArrayList;
import java.sql.*;
import java.io.IOException;

/**
 * Created by: gary at Nov 2, 2007 2:34:08 PM
 */
class BuildWorker {
    private static int ID = 0;
    private static final int MAX_ALBUMS_PER_ARTIST = 20000;
    private Thread thread = null;
    // @TODO parameterize index properties.
    final SolrClient client = SolrClient.newInstance();
    private final PreparedStatement getArtist;
    private final PreparedStatement getAlbum;
    private final PreparedStatement getTracks;
    private final ArtistIdVendor artistIds;
    private boolean done = false;
    private int id;

    private synchronized int id() {
        int x = ++ID;
        return x;
    }

    public BuildWorker(Connection con, ArtistIdVendor artistIds) throws SQLException {
        id = id();
        getArtist = con.prepareStatement("select name from Rela.Artist where artistId = ?");
        getAlbum = con.prepareStatement("select top " + MAX_ALBUMS_PER_ARTIST + " albumId, name, year, genre, discId, quality from Rela.Album where artistId = ? and albumId > ?");
//        getAlbum = con.prepareStatement("select albumId, name, year, genre, discId, quality from Rela.Album where artistId = ?");
        getTracks = con.prepareStatement("select trackId, name, extt, playOrder from Rela.Track where albumId = ? order by playOrder");
        this.artistIds = artistIds;
    }

    public synchronized void start() {
        if (thread != null) throw new RuntimeException("BuildWorker already running.");
        thread = new Thread("Index builder") {
            public void run() {
                build();
                System.out.println("DONE building index!");
                shutdown(getTracks,getAlbum,getArtist);
            }
        };
        thread.start();
    }

    private static void shutdown(Statement... statements) {
        for (Statement stmt : statements) {
            try {
                stmt.close();
            } catch (Throwable th) {
                th.printStackTrace();
            }
        }
    }

    private static long index(final SolrClient client, Item item) {
        long start = System.currentTimeMillis();
        try {
            client.index(item);
        } catch (IOException ex) {
            ex.printStackTrace();
        }
        long stop = System.currentTimeMillis();
        if (stop-start > 200)
            System.out.println("index in " + (stop-start));
        return stop - start;
    }

    boolean isDone() {
        return done;
    }

    private void build() {
        done = false;
        int artistId = artistIds.nextId();
        while (artistId > 0) {
            build(artistId);
            artistId = artistIds.nextId();
        }
        done = true;
    }

    private void build(int artistId) {
        Artist artist = getArtist(artistId);
        long start = System.currentTimeMillis();
//        System.out.println(id + " " + artist.getName() + " getting albums...");
        int startAlbum = 0;
        List<Album> albums = getAlbums(artistId,startAlbum);
        long stop = System.currentTimeMillis();
//        System.out.println(id + " " + artist.getName() + " found " + albums.size() + " in " + (stop-start));
        while (albums.size() > 0) {
            long count = 0;
            for (Album album : albums) {
                album.setArtist(artist);
                startAlbum = album.getAlbumId();
                start = System.currentTimeMillis();
                List<Track> tracks = getTracks(album.getAlbumId());
                stop = System.currentTimeMillis();
                //System.out.print("," + album.getAlbumId() + ":" + tracks.size() + ":" + (stop-start));
                for (Track track : tracks) {
                    // index track
                    track.setAlbum(album);
                    count += index(client,Item.make(track));
                }
                // index album
                count += index(client,Item.make(album));
            }
            albums = getAlbums(artistId,startAlbum);
//            System.out.println(id + " " + artist.getName() + " " + albums.size() + " albums remaining, batch took " + count);
        }
        // index artist.
        index(client,Item.make(artist));
//        System.out.println(id + " " + artist.getName() + " done.");
    }

    private List<Track> getTracks(int albumId) {
        long start = System.currentTimeMillis();
        List<Track> tracks = new ArrayList<Track>(25);
        try {
            getTracks.clearParameters();
            getTracks.setInt(1,albumId);
            ResultSet rs = getTracks.executeQuery();
            while (rs.next()) {
                Track track = new Track(rs.getInt(1),albumId,rs.getString(2),rs.getString(3),rs.getInt(4));
                tracks.add(track);
            }
            rs.close();
        } catch (SQLException ex) {
            ex.printStackTrace();
        }
        long end = System.currentTimeMillis();
        if (end-start > 100)
            System.out.println("getTracks for " + albumId + " with " + tracks.size() + " tracks in " + (end-start));
        return tracks;
    }

    private List<Album> getAlbums(int artistId, int startAlbum) {
        long start = System.currentTimeMillis();
        long sub = 0;
        List<Album> albums = new ArrayList<Album>(MAX_ALBUMS_PER_ARTIST);
        try {
            getAlbum.clearParameters();
            getAlbum.setInt(1,artistId);
            getAlbum.setInt(2,startAlbum);
            ResultSet rs = getAlbum.executeQuery();
            long xs = System.currentTimeMillis();
            while (rs.next()) {
                Album al = new Album(rs.getInt(1),artistId,rs.getString(2),rs.getString(3),rs.getString(4),rs.getString(5),rs.getInt(6));
                albums.add(al);
            }
            long xc = System.currentTimeMillis();
            sub = xc-xs;
            rs.close();
        } catch (SQLException e) {
            e.printStackTrace(); 
        }
        long end = System.currentTimeMillis();
        if (end-start > 100)
            System.out.println("getAlbums(" + artistId + "," + startAlbum + ")=>" +  albums.size() + " in " + (end-start) + ", " + sub);
        return albums;
    }

    private Artist getArtist(int artistId) {
        long start = System.currentTimeMillis();
        Artist artist = null;
        try {
            getArtist.clearParameters();
            getArtist.setInt(1,artistId);
            ResultSet rs = getArtist.executeQuery();
            if (rs.next())
                artist = new Artist(artistId, rs.getString(1));
            rs.close();
        } catch (SQLException ex) {
            ex.printStackTrace();
        }
        long end = System.currentTimeMillis();
        if (end-start > 50)
            System.out.println("getArtist(" + artistId + ") in " + (end-start));
        return artist;
    }
}
